import cProfile, pstats, io # DEBUG

print(": Registering basic commands")

class ExtendedParseStepInfo(gdb.Parameter):
	"""Controls whether to display parser stack and input preview on stepping the parse."""
	def __init__(self):
		super(ExtendedParseStepInfo, self).__init__("hammer-extended-parse-step-info", gdb.COMMAND_OBSCURE, gdb.PARAM_BOOLEAN)
		self.show_doc = "Show parser stack and input preview after hammer-parse-step:"
		#self.set_doc = "Show parser stack and input preview after hammer-parse-step:"
		self.value = True
		print(":: hammer-extended-parse-step-info")

ExtendedParseStepInfo()

class HammerParserBacktrace(gdb.Command):
	def __init__(self):
		super(HammerParserBacktrace, self).__init__ ("hammer-parser-backtrace", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-backtrace")

	def invoke(self, arg, from_tty):
		parserstack = top_level_parse.peek_parserstack().p_stack
		args = gdb.string_to_argv(arg)
		if len(args) < 1:
			maxsize = len(parserstack)
		else:
			try:
				maxsize = int(args[0])
				if maxsize < 1:
					raise ValueError
			except ValueError:
				maxsize = len(parserstack)
				print("Argument must be a positive integer")

		parser_name = top_level_parse.h_do_parse_parser.name or "(Parser type resolution pending)" # TODO: should we just look at the vtable and name the parser in h_do_parse?

		print("[" + str(hex(top_level_parse.h_do_parse_parser.address)) + "] " + parser_name + " [current]") #TODO: GUI widget should reflect this
		print(" ")
		depth = min(len(parserstack), maxsize)
		if depth > 0: # if stack not empty
			# unsure what the idiomatic python is for handling negative indices starting with -1,
			# but this addition is to avoid off-by-one errors
			index = -(depth+1)
			for p in parserstack[-1:index:-1]:
				print("[" + str(hex(p.address)) + "] " + p.name) # TODO: errors in perform_lowlevel_parse, if p.name is None
			if depth < len(parserstack):
				print("[...]")

HammerParserBacktrace()

class HammerParserMemUse(gdb.Command):
	def __init__(self):
		super(HammerParserMemUse, self).__init__("hammer-parser-mem-use", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-mem-use")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)
		if len(args) < 1:
			print("Usage: hammer-parser-mem-use <address>")
			return

		parser_addr = args[0]
		try:
			parser_addr_int = int(parser_addr, 16)
			parser_obj = top_level_parse.parser_by_address(parser_addr_int)
			if parser_obj is not None:
				print(parser_obj.bytes_used)
		except ValueError:
			print("Address needs to be a hexadecimal number")

HammerParserMemUse()

class HammerParserMemUseName(gdb.Command):
	def __init__(self):
		super(HammerParserMemUseName, self).__init__("hammer-parser-mem-use-name", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-mem-use-name")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)
		if len(args) < 1:
			print("Usage: hammer-parser-mem-use-name <name>")
			return

		parser_name = args[0]
		parser_objs = top_level_parse.parsers_by_name(parser_name)
		if parser_objs is not None:
			for p in parser_objs:
				print((p.name, hex(p.address), p.bytes_used))

HammerParserMemUseName()

class HammerParserTopSingleArenaMem(gdb.Command):
	def __init__(self):
		super(HammerParserTopSingleArenaMem, self).__init__("hammer-parser-top-single-arena-mem", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-top-single-arena-mem")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)

		p = top_level_parse.get_parser_top_per_arena_mem()
		print((p.name, hex(p.address), p.bytes_used))

HammerParserTopSingleArenaMem()

class HammerParserTopTotalArenaMem(gdb.Command):
	def __init__(self):
		super(HammerParserTopTotalArenaMem, self).__init__("hammer-parser-top-total-arena-mem", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-top-total-arena-mem")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)

		p = top_level_parse.get_parser_top_total_arena_mem()
		print((p.name, hex(p.address), p.bytes_used))
		total_mem_use = p.get_arenasum()
		print("Total: " + str(total_mem_use) + " bytes")

HammerParserTopTotalArenaMem()

# TODO: average memory use, per arena and total

class HammerParserPreviewInput(gdb.Command):
	def __init__(self):
		super(HammerParserPreviewInput, self).__init__("hammer-parser-preview-input", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-preview-input")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)

		print(top_level_parse.input_chunk.replace("\x00", ".")) # Avoid embedded null error

HammerParserPreviewInput()

class HammerParserAverageMem(gdb.Command):
	def __init__(self):
		super(HammerParserAverageMem, self).__init__("hammer-parser-average-mem", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-average-mem")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)

		mem = top_level_parse.get_avg_mem_use_per_arena()
		print("Bytes used on average in each arena:")
		print(mem)

HammerParserAverageMem()

class HammerParserCurrentEnv(gdb.Command):
	def __init__(self):
		super(HammerParserCurrentEnv, self).__init__("hammer-parser-current-env", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-current-env")

	def invoke(self, arg, from_tty):
		p = top_level_parse.h_do_parse_parser
		p_env = top_level_parse.parser_decombinator.decompose_parser(p, top_level_parse) #TODO: parser -> env mapping function in top_level_parse
		print(type(p_env).__name__ + " - " + str(p_env)) # TODO: consistency with GUI

HammerParserCurrentEnv()

class HammerParserDumpMemoryStats(gdb.Command):
	def __init__(self):
		super(HammerParserDumpMemoryStats, self).__init__("hammer-parser-dump-memory-stats", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-dump-memory-stats")

	def invoke(self, arg, from_tty):
		stats = top_level_parse.dump_memory_stats()
		print(stats)

HammerParserDumpMemoryStats()

# When the hammer-parse-step-to-result command is executed, top_level_parse
# marks the stack location whose result is of interest of the user.
# The stack grows during parsing, and eventually gets popped back to the marked location.
# At the time the associated h_do_parse() returns, the partial AST, constructed from the bottom up, will be ready
# E.g. hammer-parse-step-to-parse-result 5 will stop after the 6th parser from the top of the parser backtrace has been applied
# hammer-parse-step-after-apply -5 will stop after the 5th parser from the bottom of the backtrace has been applied.
# A parameter of 0 will result in stepping to when the application of the "current" parser finishes.
# Passing parser here is not strictly necessary, but probably helps debug off-by-one errors
class HammerParseStepToResult(FlowControlWithPrint):
	def __init__(self):
		super(HammerParseStepToResult, self).__init__("hammer-parse-step-to-result", gdb.COMMAND_OBSCURE)
		print(":: hammer-parse-step-to-result")

	def invoke(self, arg, from_tty):
		args = gdb.string_to_argv(arg)
		index = 0
		if len(args) < 1:
			print("No argument given. Execution will be stopped after the current parser being applied returns its result.")
		else:
			try:
				index = int(args[0])
			except ValueError:
				print("Argument needs to be an integer. Execution will be stopped after the current parser being applied returns its result.")

		top_level_parse.setup_ast_stack_index(index)
		# HDoParseRetBreakpoint sets hammer_step_counter to 1 after capturing the AST subtree, so this will stop at the first h_do_parse invocation afterwards
		if gdb.selected_inferior().pid > 0:
			gdb.execute("continue")

		top_level_parse.clear_ast_stack_index()

		self.conditionally_print_backtrace()

HammerParseStepToResult()

# When invoked without argument, it'll attempt to print the AST held by ast_manager. This may or may not be outdated.
# When invoked with an integer argument, it'll remember the selection, and print the AST when the HParseResult is returned by h_do_parse.
class HammerParserPrintAST(gdb.Command):
	def __init__(self):
		super(HammerParserPrintAST, self).__init__("hammer-parser-print-ast", gdb.COMMAND_OBSCURE)
		print(":: hammer-parser-print-ast")

	def invoke(self, arg, from_tty):
		ast_manager.print_ast()

HammerParserPrintAST()

# Step until the "current" parser is applied and returns the HParsedToken, stopping in h_do_parse
class HammerParseApply(FlowControlWithPrint):
	def __init__(self):
		super(HammerParseApply, self).__init__("hammer-parse-apply", gdb.COMMAND_OBSCURE)
		print(":: hammer-parse-apply")

	def invoke(self, arg, from_tty):
		#profiler = cProfile.Profile() # DEBUG
		#profiler.enable() # DEBUG
		with cProfile.Profile() as profiler: # DEBUG
			top_level_parse.setup_ast_stack_index(0) #TODO: is it a problem if this command overwrites it? would it better to use a convenience variable, as with parse-step?
			showmap_old = top_level_parse.show_input_map_after_apply
			top_level_parse.show_input_map_after_apply = False

			# HDoParseRetBreakpoint sets hammer_step_counter to 1 after capturing the AST subtree, so this will stop at the first h_do_parse invocation afterwards
			if gdb.selected_inferior().pid > 0:
				gdb.execute("continue")

			top_level_parse.show_input_map_after_apply = showmap_old
			top_level_parse.clear_ast_stack_index()

			self.conditionally_print_backtrace()
		profiler.print_stats(sort='tottime') # DEBUG
		#profiler.disable() # DEBUG
		#s = io.StringIO() # DEBUG
		#sortby = SortKey.CUMULATIVE # DEBUG
		#ps = pstats.Stats(pr, stream=s).sort_stats(sortby) # DEBUG
		#ps.print_stats() # DEBUG
		#print(s.getvalue()) # DEBUG

HammerParseApply()

class HammerParseApplyAndShowAST(FlowControlWithPrint):
	def __init__(self):
		super(HammerParseApplyAndShowAST, self).__init__("hammer-parse-apply-and-show-ast", gdb.COMMAND_OBSCURE)
		print(":: hammer-parse-apply-and-show-ast")

	def invoke(self, arg, from_tty):
		#profiler = cProfile.Profile() # DEBUG
		#profiler.enable() # DEBUG
		with cProfile.Profile() as profiler: # DEBUG
			top_level_parse.setup_ast_stack_index(0) #TODO: is it a problem if this command overwrites it? would it better to use a convenience variable, as with parse-step?
			showmap_old = top_level_parse.show_input_map_after_apply
			top_level_parse.show_input_map_after_apply = True

			# HDoParseRetBreakpoint sets hammer_step_counter to 1 after capturing the AST subtree, so this will stop at the first h_do_parse invocation afterwards
			if gdb.selected_inferior().pid > 0:
				gdb.execute("continue")

			top_level_parse.show_input_map_after_apply = showmap_old
			top_level_parse.clear_ast_stack_index()

			self.conditionally_print_backtrace()
		profiler.print_stats(sort='tottime') # DEBUG
		#profiler.disable() # DEBUG
		#s = io.StringIO() # DEBUG
		#sortby = SortKey.CUMULATIVE # DEBUG
		#ps = pstats.Stats(pr, stream=s).sort_stats(sortby) # DEBUG
		#ps.print_stats() # DEBUG
		#print(s.getvalue()) # DEBUG

HammerParseApplyAndShowAST()
